require 'socket'
require 'thread'
require 'fcntl'
require 'druby4_helper'

=begin
DRuby (not to be confused with dRuby, written by Masatoshi SEKI)
(C) Copyright 2001 Paul Brannan (cout at rm-f.net)

DRuby is a set of classes for providing distributed object support to a
Ruby program.  You may distribute and/or modify it under the same terms as
Ruby (see http://www.ruby-lang.org/en/LICENSE.txt).  Example:

Client 
------
client = DRuby::Client.new('localhost', 4242)
obj = client.resolve("foo")
puts obj.foo("foo")
obj.oneway(:foo, "foo")

Server
------
class Foo
    def foo(x); return x; end
end
obj = Foo.new
server = DRuby::Server.new('localhost', 4242)
server.bind(obj, "foo")
server.thread.join

You can do all sorts of cool things with DRuby, including passing blocks to
the functions, throwing exceptions and propogating them from server to
client, and more.  Unlike CORBA, where you must create an interface
definition and strictly adhere to it, DRuby uses marshalling, so you can
use almost any object with it.  But, alas, it is not as powerful as CORBA.

On a fast machine, you should expect around 7000 messages per second with
normal method calls, up to 10000 messages per second with oneway calls with
sync, and up to 40000 messages per second for oneway calls without sync.

The DRuby message format is broken into 3 components:
    [ msg_type, obj_id, message ]
For each msg_type, there is a specific format to the message.  Additionally,
certain msg_types are only valid when being sent to the server, and others
are valid only when being sent back to the client.  Here is a summary:

msg_type         send to     meaning of obj_id       msg format
----------------------------------------------------------------------------
REQUEST          server      obj to talk to          [:method, *args]
REQUEST_BLOCK    server      obj to talk to          [:method, *args]
ONEWAY           server      obj to talk to          [:method, *args]
ONEWAY_SYNC      server      obj to talk to          [:method, *args] 
RETVAL           client      always 0                retval
EXCEPTION        client      always 0                $!
YIELD            client      always 0                [value, value, ...]
SYNC             either      0=request, 1=response   nil
NULL_MSG         either      always 0                n/a

BUGS:
On a 2.2 kernel, oneway calls without sync is very slow.
=end

module DRuby

public

    # The DRuby server class.  Like its drb equivalent, this class spawns off
    # a new thread which processes requests, allowing the server to do other
    # things while it is doing processing for a distributed object.  This
    # means, though, that all objects used with DRuby must be thread-safe.
    class Server

    public
        attr_reader :host, :port, :obj, :thread

        # Start a server on the given host and port.  An acceptor may be
        # specified to accept/reject connections; it should be a proc that
        # takes a Socket as an argument and returns true or false.
        def initialize(host, port, acceptor=nil, debug=false)
            @host = host
            @port = port
            @mutex = Mutex.new
            @debug = debug
            @resolve_server = Resolve_Server.new
            @resolve_obj = Resolve_Obj.new(@resolve_server)
            @resolve_server.register(@resolve_obj)

            @thread = Thread.new do
                server = TCPServer.new(@host, @port)
                while(socket = server.accept)
                    puts "Got a connection" if @debug
                    if acceptor then
                        if !acceptor.call(socket) then
                            socket.close
                            next
                        end
                    end
                    puts "Accepted the connection" if @debug
                    socket.setsockopt(Socket::SOL_TCP, Socket::TCP_NODELAY, 1)
                    socket.fcntl(Fcntl::F_SETFL, Fcntl::O_NONBLOCK)
                    socket.sync = true
                    session = Session.new(socket)
                    session.set_nonblock(true)
                    Thread.new(socket) do |socket|
                        begin
                            server_loop(session)
                        rescue Exception
                            DRuby::print_exception($!) if @debug
                        end
                        puts "Connection closed" if @debug
                    end
                end
            end
        end

        # Register an object with the server.  The object will be given an
        # id of @next_id, and @next_id will be incremented.  We could use the
        # object's real id, but this is insecure.  The supplied object must
        # be thread-safe.
        def create_reference(obj)
            @mutex.synchronize do
                id = @resolve_server.register(obj)
                Object_Reference.new(id) #return
            end
        end

        # Finds an object in linear time and unregister it.  If you need a
        # faster function, write one and send me a patch.
        # Be careful with this function, because the client may not know the
        # object has gone away.
        def delete_reference(obj)
            @mutex.synchronize do
                @resolve_server.unregister(obj)
            end
            nil #return
        end

        # Register an object with the server and bind it to name.
        def bind(obj, name)
            id = @resolve_server.register(obj)
            @resolve_server.bind(name, id)
            nil #return
        end

        # This keeps the client from seeing our objects when they call inspect
        alias_method :__inspect__, :inspect
        def inspect()
            return ""
        end

    private
        if false then # the following functions are implemented in C:

        # The server_loop function is the guts of the server.  It takes in
        # requests from the client and forwards them to already-registered
        # objects.
        def server_loop(session)
        end

        end # if false
    end

    # The DRuby client class.  A DRuby server must be started on the given
    # host and port before instantiating a DRuby client.
    class Client
        attr_reader :host, :port
        
        # Connect a client to the server on the given host and port.  The
        # user can specify sync=false to turn off synchronization and get a
        # 20% speed boost.
        def initialize(host, port, sync=true)
            @host = host
            @port = port
            @server = TCPSocket.open(@host, @port)
            @server.sync = true
            @server.setsockopt(Socket::SOL_TCP, Socket::TCP_NODELAY, 1)
            @server.fcntl(Fcntl::F_SETFL, Fcntl::O_NONBLOCK)
            @session = Session.new(@server)
            @session.set_nonblock(true)
            @mutex = sync ? Mutex.new : Null_Mutex.new
            @resolve_obj = Proxy_Object.new(@session, @mutex, 0)
        end

        # Given a string, return a proxy object that will forward requests
        # for an object on the server with that name.
        def resolve(object_name)
            @mutex.synchronize do
                object_id = @resolve_obj.resolve(object_name)
                return Proxy_Object.new(@session, @mutex, object_id)
            end
        end
    end

private

    # In case the user does not want synchronization.
    class Null_Mutex
        def synchronize; yield; end
        def lock; end
        def unlock; end
    end

    # All the special functions we have to keep track of
    class Functions
        GOOD = [
            :inspect, :class_variables, :instance_eval, :instance_variables,
            :to_a, :to_s
        ]

        BAD = [
            :clone, :dup, :display
        ]

        METHOD = [
            :methods, :private_methods, :protected_methods, :public_methods,
            :singleton_methods
        ]

        RESPOND = [
            [ :method,      "raise NameError" ],
            [ :respond_to?, "false" ]
        ]
    end

    # A DRuby::Object_Reference is created on the server side to represent an
    # object in the system.  It can be returned from a server object to a
    # client object, at which point it is converted into a DRuby::Proxy_Object.
    class Object_Reference
        attr_reader :object_id

        def initialize(object_id)
            @object_id = object_id
        end
    end

    # A DRuby::Object acts as a proxy; it forwards most methods to the server
    # for execution.  When you make calls to a DRuby server, you will be
    # making the calls through a Proxy_Object.
    class Proxy_Object

        if false then # the following functions are implemented in C:

        # The method_missing function is called for any method that is not
        # defined on the client side.  It will forward requests to the server
        # for processing, and can iterate through a block, raise an exception,
        # or return a value.
        def method_missing(function, *args)
        end

        # The oneway function is called to make a oneway call to the server
        # without synchronization.
        def onweway(function, *args)
        end

        # The oneway_sync function is called to make a oneway call to the
        # server with synchronization (the server will return a null message
        # to the client before it begins processing).  This is slightly safer
        # than a normal oneway call, but it is slower (except on a linux 2.2
        # kernel; see the bug list above).
        def oneway_sync(function, *args)
        end
        
        # The sync function will synchronize with the server.  It sends a sync
        # request and waits for a response.
        def sync()
        end

        end # if false

        # Make sure certain methods get passed down the wire.
        Functions::GOOD.each do |method|
            eval %{
                def #{method}(*args)
                      method_missing(:#{method}, *args) #return
                end
            }
        end

        # And make sure others never get called.
        Functions::BAD.each do |method|
            eval %{
                def #{method}(*args)
                    raise(NameError,
                        "undefined method `#{method}' for " +
                        "\#<#{self.class}:#{self.id}>")
                end
            }
        end

        # And remove these function names from any method lists that get
        # returned; there's nothing we can do about people who decide to
        # return them from other functions.
        Functions::METHOD.each do |method|
            eval %{
                def #{method}(*args)
                    retval = method_missing(:#{method}, *args)
                    retval.each do |item|
                        Functions::BAD.each do |bad|
                            retval.delete(bad.to_s)
                        end
                    end
                    retval #return
                end
            }
        end

        # Same here, except don't let the call go through in the first place.
        Functions::RESPOND.each do |method, action|
            eval %{
                def #{method}(arg, *args)
                    Functions::BAD.each do |bad|
                        if arg === bad.to_s then
                            return eval("#{action}")
                        end
                    end
                    method_missing(:#{method}, arg, *args) #return
                end
            }
        end

    end

    # The Resolve_Server class registers objects for the server.  You will
    # never have to use this class directly.
    class Resolve_Server
        def initialize
            @next_id = 0
            @unused_ids = Array.new
            @id_to_object = Hash.new
            @name_to_id = Hash.new
        end

        def register(obj)
            if @next_id >= Session::MAX_ID then
                if @unused_ids.size == 0 then
                    raise "Object limit exceeded"
                else
                    id = @unused_ids.pop
                end
            end
            @id_to_object[@next_id] = obj
            old_id = @next_id
            @next_id = @next_id.succ()
            old_id #return
        end

        def get_object(object_id)
            @id_to_object[object_id] #return
        end

        def unregister(obj)
            delete_obj_from_array_private(@id_to_object, obj)
        end

        def bind(name, id)
            @name_to_id[name] = id
        end

        def resolve(name)
            @name_to_id[name] #return
        end

        def delete_obj_from_array_private(array, obj)
            index = array.index(obj)
            array[index] = nil unless index == nil
        end
    end

    # The Resolve_Obj class handles resolve requests for the client.  It is
    # a special DRuby object with an object id of 0.  You will never have to
    # make calls on it directly, but will instead make calls on it through
    # the Client object.
    class Resolve_Obj
        def initialize(resolve_server)
            @resolve_server = resolve_server
        end

        def resolve(name)
            @resolve_server.resolve(name) #return
        end
    end

    # Print an exception to the screen.  This is necessary, because Ruby does
    # not give us access to its error_print function from within Ruby.
    def self.print_exception(exc)
        first = true
        $!.backtrace.each do |bt|
            if first then
                puts "#{bt}: #{$!} (#{$!.message})"
            else
                puts "\tfrom #{bt}"
            end
            first = false
        end
    end

    # The Sesssion class is defined in druby4_helper.so.  You should never have
    # to use it directly.

end
